unit TomasCollectionForm;

(***********************************************************************)
(*  This software is copyleft software;                                *)
(*  You may distribute or change the software as you wish, but         *)
(*  this header must always be present.                                *)
(*  No one is allowed to claim copyright on this source code.          *)
(*  All precompiled software using this unit should be accompanied     *)
(*  by this source code.                                               *)
(*                                                                     *)
(*  Original contents on http://www.tomasweb.com                       *)
(*  Contact: Hans P.M. Veeke         H.P.M.Veeke@wbmt.tudelft.nl       *)
(***********************************************************************)

(***********************************************************************)
(* Update history                                                      *)
(* 17-04-2001: zero-divide bug solved for constant values              *)
(* 14-08-2000: sethistogramtype corrected                              *)
(*             'freeing' a collection corrected, because if was        *)
(*             not automatically removed from a figure                 *)
(* Original version: 19-06-2000                                        *)
(***********************************************************************)
interface

uses
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  Menus,
  ComCtrls,
  OleCtrls,
  vcfi,
  StdCtrls,
  Buttons,
  ExtCtrls;

type

  CollectionViewType = (Bar, Graph, Histogram,None);
  HistogramViewType = (HDiscrete,HCumulative,Both);

  MinMaxRecord =
    Record
    MinX: Double;
    MaxX: Double;
    MinY: Double;
    MaxY: Double;
    End;

  TomasCollectionData = class;
  TomasCollectionFigure = class;
  TCollectionStatusBar = class;

  TCollectionForm = class(TForm)
    MainMenu1: TMainMenu;
    FigureItem: TMenuItem;
    BarItem: TMenuItem;
    GraphItem: TMenuItem;
    HistogramItem: TMenuItem;
    AxisItem: TMenuItem;
    XYPanel: TPanel;
    XValuesGroup: TGroupBox;
    XMinEdit: TEdit;
    Label1: TLabel;
    XRangeLabel: TLabel;
    XRangeEdit: TEdit;
    ClassWidthEdit: TEdit;
    ClassLabel: TLabel;
    YValuesGroup: TGroupBox;
    YMinedit: TEdit;
    Label2: TLabel;
    YMaxEdit: TEdit;
    Label3: TLabel;
    ValuesOKButton: TBitBtn;
    ValuesCancelButton: TBitBtn;
    NoneItem: TMenuItem;
    HistogramTypeGroup: TRadioGroup;
    PrintFigureItem: TMenuItem;
    PrintDialog: TPrintDialog;
    procedure GoCreateCollectionForm(Sender: TObject);
    procedure SetBarType(Sender: TObject);
    procedure SetGraphType(Sender: TObject);
    procedure SetHistogramType(Sender: TObject);
    procedure AxisItemClick(Sender: TObject);
    procedure HideXYValues(Sender: TObject);
    procedure GoPaintCollectionForm(Sender: TObject);
    procedure GoAcceptValues(Sender: TObject);
    procedure GoResizeForm(Sender: TObject);
    procedure NoneItemClick(Sender: TObject);
    procedure CheckCloseFigure(Sender: TObject; var Action: TCloseAction);
    procedure GoPrintFigure(Sender: TObject);
  private
    { Private declarations }
    Figure: TomasCollectionFigure;
    StatusBarList: TList;
    FRect: TRect;
    Procedure ReDrawFigure;
  public
    { Public declarations }

  end;

  PCollectionValue = ^TCollectionValue;
  TCollectionValue =
    Record
    XValue: Double;
    YValue: Double;
    End;

  TomasCollectionData = class
    Private
    FValues: TList;
    FClasses: TList;
    FSigmaY: Double;
    FSigmaY2: Double;
    FNrValues: Integer;
    FMinMax: MinMaxRecord;
    FColor: TColor;
    FStatusBar: TCollectionStatusBar;
    FFigure: TomasCollectionFigure;
    FCSV: String;
    XFrom: Integer;
    YFrom: integer;
    Xto: integer;
    Yto: Integer;
    LastValueIndex: Integer;
    Quantiles: Array[1..200] Of Integer;
    QClassWidth: Double;
    QLower: Double;
    Procedure FillHistogram;
    Published
    Constructor Create;
    Destructor Destroy; override;
    Procedure Collect(x,y: Double);
    Procedure SetColor(Color: TColor);
    procedure SetCSVFile(CSV: String);
    Property NrOfValues: Integer read FNrValues;
    Procedure GetValue(i: Integer; Var X,Y: Double);
    Procedure ClearValues;
    Property CColor: TColor read FColor write SetColor;
    Property Figure: TomasCollectionFigure read FFigure;
    property CSVFile: String read FCSV write SetCSVFile;
    Function Mean: Double;
    function Minimum : double;
    function Maximum : double;
    Function StandardDeviation: Double;
    Function Quantile(Quantile: Double): Double;
    Procedure SaveToCSV(CSV,Title: String);
    Procedure RefreshData;
    End;

  TomasCollectionFigure = class
  Private
    FTitle: String;
    FForm: TCollectionForm;
    FView: CollectionViewType;
    FDataList: TList;   {lists with FValues of each collection}
    FMinMax: MinmaxRecord;
    FRange: Double;
    FNrClasses: Integer;
    FClassWidth: Double;
    FHistogramView: HistogramViewType;
    FValueAdded: Boolean;  {to prevent complete redraw when only 1 value is added}
    FLastValues: TList;    {sorted list for Bar presentation}
    FFirstCall: Boolean;
    Function GetNextBarValue(Var Val: PCollectionValue): TomasCollectionData;
  Published
    Constructor Create(Figname: String);
    Destructor Destroy; Override;
    Procedure SetViewType(VType: CollectionViewType);
    Property ViewType: CollectionViewType read FView write SetViewType;
    Property Title: String read FTitle;
    Function NrOfCollections: Integer;
    Procedure AddCollection(Collection: TomasCollectionData; Title: String);
    Procedure RemoveCollection(Collection: TomasCollectionData);
    Procedure SetFValueAdded;
    Procedure Repaint;
    procedure GoDrawPanel(StatusBar: TStatusBar;
     Panel: TStatusPanel; const Rect: TRect);
    Procedure MakeHistograms;
    procedure SaveToSettingsFile(Var F: TextFile);
    Procedure ReadFromSettingsFile(Var F: textFile);
  End;

  PHistDataLink = ^HistDataLink;
  HistDataLink =
    Record
    ClassLB: Double;
    NrObservations: Integer;
    Percentage: Double;
    Color: TColor;
    End;



  TCollectionStatusBar = class(TStatusBar)
    Public
    Collection: TomasCollectionData;
    End;

Function ViewTypetoString(VType: CollectionViewType): String;

var
  CollectionForm: TCollectionForm;

implementation

{$R *.DFM}

Const
  NEVER: Double = 1.7E+307;


(******************************************************************)
(*                        supporting methods                      *)
(******************************************************************)

Function ViewTypetoString(VType: CollectionViewType): String;
  Begin
  Case VType Of
    Bar: Result:='Bar';
    Graph: Result:='Graph';
    Histogram: Result:='Histogram';
    Else Result:='None';
    End;
  End;  {ViewTypeToString}

Function StringToViewType(VString: String): CollectionViewType;
  Begin
  If UpperCase(VString) = 'BAR' Then
    Begin
    Result:=Bar;
    Exit;
    End;
   If UpperCase(VString) = 'GRAPH' Then
    Begin
    Result:=Graph;
    Exit;
    End;
  If UpperCase(VString) = 'HISTOGRAM' Then
    Begin
    Result:=Histogram;
    Exit;
    End;
  Result:=None;
  End;


(********************************************************************)
(*                  Methods of TomasCollectionData                  *)
(********************************************************************)

Constructor TomasCollectionData.Create;
  Var
    i: Integer;
  Begin
  FValues:=TList.Create;
  FClasses:=TList.Create;
  FMinMax.MinX:=Never;
  FMinMax.MaxX:=-Never;
  FMinMax.MinY:=Never;
  FMinMax.MaxY:=-Never;
  FFigure:=Nil;
  FColor:=clRed;
  FSigmaY:=0;
  FSigmaY2:=0;
  FNrValues:=0;
  For i:=1 To 200 Do
    Begin
    Quantiles[i]:=0;
    End;
  QClassWidth:=-1;
  FCSV:='';
  End;

Destructor TomasCollectionData.Destroy;
  Var
    DValue: PCollectionValue;
  Begin
  While FValues.Count > 0 Do
    Begin
    DValue:=FValues.Last;
    FValues.Remove(DValue);
    Dispose(DValue);
    End;
  FValues.Free;
  If Assigned(FFigure) Then
    Begin
    FFigure.FDataList.Remove(Self);
    End;
  End;

Procedure TomasCollectionData.Collect(x,y: Double);
  Var
    NewValue: PCollectionValue;
    i: Integer;
    j: Integer;
    Delta: Double;
    HistData: PHistDataLink;
  Begin
  New(NewValue);
  NewValue.XValue:=x;
  NewValue.YValue:=y;
  FValues.Add(NewValue);
  FSigmaY:=FSigmaY + y;
  FSigmaY2:=FSigmaY2 + y * y;
  Inc(FNrValues);
  If FMinMax.MinX > x Then
    Begin
    FMinMax.MinX:=x;
    End;
  If FMinMax.MaxX < x Then
    Begin
    FMinMax.MaxX:=x;
    End;
  If FMinMax.MinY > y Then
    Begin
    FMinMax.MinY:=y;
    End;
  If FMinMax.MaxY < y Then
    Begin
    FMinMax.MaxY:=y;
    End;
  If (FNrValues = 100) And (QClassWidth = -1) Then
    Begin
    QClassWidth:=(FMinMax.MaxY - FMinmax.MinY)/100;
    QLower:=FMinMax.MinY - 50 * QClassWidth;
    For j:=0 To FValues.Count - 1 Do
      Begin
      NewValue:=FValues.Items[j];
      If QClassWidth <> 0 Then
        Begin
        i:= Trunc((NewValue.YValue - QLower)/QClassWidth) + 1;
        End
      Else
        Begin
        i:=1;
        End;
      Inc(Quantiles[i]);
      End;
    End;
  If FNrValues > 100 Then
    Begin
    If QClassWidth > 1e-8 Then
      Begin
      i:= Trunc((NewValue.YValue - QLower)/QClassWidth) + 1;
      End
    Else
      Begin
      i:=1;
      End;
    If i < 1 Then
      Begin
      i:=1;
      End;
    If i> 200 Then
      Begin
      i:=200;
      End;
    Inc(Quantiles[i]);
    End;
  If FFigure <> Nil Then
    Begin
    If (FClasses.Count > 0) Then
      Begin
      If FFigure.FClassWidth <> 0 Then
        Begin
        j:=Trunc((Y - FFigure.FMinMax.MinY)/FFigure.FClassWidth) + 1;
        End
      Else
        Begin
        j:=1;
        End;
      If j <= 0 Then
        Begin
        j:=0;
        End;
      If j > FFigure.FNrClasses + 1 Then
        Begin
        j:=FFigure.FNrClasses + 1;
        End;
      HistData:=FClasses[j];
      Inc(HistData^.NrObservations);
      End;
    RefreshData;
    If FFigure.FView <> Histogram Then
      Begin
      If X > FFigure.FMinMax.MaxX Then
        Begin
        Delta:=(FFigure.FMinMax.MaxX - FFigure.FMinMax.MinX) / 2;
        FFigure.FMinMax.MinX:=FFigure.FMinMax.MinX + Delta;
        FFigure.FMinmax.MaxX:=FFigure.FMinMax.MaxX + Delta;
        FFigure.FForm.RedrawFigure;
        End
      Else
        Begin
        FFigure.SetFValueAdded;
        FFigure.Repaint;
        End
      End
    Else
      begin
      FFigure.Repaint;
      End;
    End
  End;  {TomasCollectionData.Collect}

Procedure TomasCollectionData.ClearValues;
  Var
    RValue: PCollectionValue;
  Begin
  While FValues.Count > 0 Do
    Begin
    RValue:=FValues.Items[0];
    FValues.Remove(RValue);
    Dispose(RValue);
    End;
  FValues.Pack;
  FNrValues:=0;
  End;

Procedure TomasCollectionData.SetColor(Color: TColor);
  Begin
  FColor:=Color;
  If Assigned(FStatusBar) Then
    Begin
    FStatusBar.Repaint;
    RefreshData;
    If FFigure <> Nil Then
      Begin
      FFigure.Repaint;
      End;
    End;
  End;  {TomasCollectionData.SetColor}

procedure TomasCollectionData.SetCSVFile(CSV: String);
  Begin
  FCSV:=CSV;
  End;   {TomasCollectionDat.SetCSVFile}

Function TomasCollectionData.Mean: Double;
  Begin
  If FNrValues = 0 Then
    Begin
    Result:=0;
    End
  Else
    Begin
    Result:=FSigmaY/FNrValues;
    End;
  End;  {TomasCollectionData.Mean}

Function TomasCollectionData.Minimum: Double;
  Begin
  If FNrValues = 0 Then
    Begin
    Result:=0;
    End
  Else
    Begin
    Result:=FMinMax.MinY;
    End;
  End;  {TomasCollectionData.Minimum}

Function TomasCollectionData.Maximum: Double;
  Begin
  If FNrValues = 0 Then
    Begin
    Result:=0;
    End
  Else
    Begin
    Result:=FMinMax.MaxY;
    End;
  End;  {TomasCollectionData.Maximum}

Function TomasCollectionData.StandardDeviation: Double;
  var
    V: Double;
  Begin
  If FNrValues <= 1 Then
    Begin
    Result:=0;
    End
  Else
    Begin
    V:=FSigmaY2 - (FSigmaY * FSigmaY)/FNrValues;
    If V <= 0 Then
      Result:=0
    Else
      Result:=Sqrt((FSigmaY2 - (FSigmaY * FSigmaY)/FNrValues )/(FNrValues - 1));
    End;
  End;  {TomasCollectionData.StandardDeviation}

Function TomasCollectionData.Quantile(Quantile: Double): Double;
  Var
    x: Double;
    i: Integer;
    j: Integer;
    k: Integer;
  Begin
  If FNrValues < 100 Then
    Begin
    Result:=0;
    End
  Else
    Begin
    x:=Quantile * FNrValues / 100.0;
    i:=Trunc(x);
    j:=FNrValues;
    k:=201;
    While j > i Do
      Begin
      Dec(k);
      j:=j - Quantiles[k];
      End;
    Result:=QLower + (k - 1 + x - i) * QClassWidth;
    End;
  End;  {TomasCollectionData.Quantile}

Procedure TomasCollectionData.RefreshData;
  Begin
  If Assigned(FStatusBar) Then
    Begin
    FStatusBar.Panels[2].Text:='Values ' + IntToStr(FNrValues);
    FStatusBar.Panels[3].Text:='Avg: ' + FormatFloat('0.0',Mean);
    FStatusBar.Panels[4].Text:='Std: ' + FormatFloat('0.0',StandardDeviation);
    FStatusBar.Panels[5].Text:='95%: ' + FormatFloat('0.0',Quantile(95));
    FStatusBar.Panels[6].Text:='Max: ' + FormatFloat('0.0',FMinMax.MaxY);
    End;
  End;

Procedure TomasCollectionData.SaveToCSV(CSV,Title: String);
  Var
    CSVF: TextFile;
    Value: PCollectionValue;
    P: Integer;
    i: Integer;
  Begin
  If CSV <> '' Then
    Begin
    P:=Pos('.',CSV);
    If P = 0 Then
      begin
      CSV:=CSV + '.csv';
      End;
    AssignFile(CSVF,CSV);
    ReWrite(CSVF);
    WriteLn(CSVF,'X,' + Title);
    For i:=0 To FValues.Count -1 Do
      Begin
      Value:=FValues.Items[i];
      WriteLn(CSVF,FormatFloat('0.0',Value.XValue) + ','
       + FormatFloat('0.0',Value.YValue));
      End;
    CloseFile(CSVF);
    End;
  End;

Procedure TomasCollectionData.GetValue(i: Integer; Var X,Y: Double);
  Var
    Val: PCollectionValue;
  Begin
  If (i >= 0) And (i <= FValues.Count - 1) Then
    Begin
    Val:=FValues.Items[i];
    X:=Val.XValue;
    Y:=Val.YValue;
    End
  Else
    Begin
    X:=0;
    Y:=0;
    End;   
  End;

Procedure TomasCollectionData.FillHistogram;
  Var
    Value: PCollectionValue;
    HistData: PHistDataLink;
    i: integer;
    j: Integer;
  Begin
  If FValues.Count > 0 Then
    begin
    For i:=0 To Fvalues.Count - 1 Do
      Begin
      Value:=FValues.Items[i];
      If Figure.FClassWidth <> 0 Then
        Begin
        j:=Trunc((Value^.YValue - Figure.FMinMax.MinY)/Figure.FClassWidth) + 1;
        End
      Else
        Begin
        j:=1;
        End; 
      If j <= 0 Then
        Begin
        j:=0;
        End;
      If j > Figure.FNrClasses + 1 Then
        Begin
        j:=Figure.FNrClasses + 1;
        End;
      HistData:=FClasses[j];
      Inc(HistData^.NrObservations);
      End;
    End;
  End;

(****************************************************************)
(*                  Methods of TomasCollectionFigure            *)
(****************************************************************)

Constructor TomasCollectionFigure.Create(FigName: String);
  begin
  Application.CreateForm(TCollectionForm,FForm);
  FTitle:=FigName;
  FForm.Caption:=FigName;
  FForm.Figure:=Self;
  FView:=None;
  FDataList:=TList.Create;
  FMinMax.MinX:=0;
  FMinMax.MaxX:=1000;
  FMinMax.MinY:=0;
  FMinMax.MaxY:=100;
  FRange:=200;
  FNrClasses:=20;
  FClassWidth:=10;
  FHistogramView:=HDiscrete;
  FForm.Show;
  FForm.BringToFront;
  FValueAdded:=FALSE;
  FLastValues:=TList.Create;
  FFirstCall:=FALSE;
  End;  {TomasCollectionFigure.Create}

Destructor TomasCollectionFigure.Destroy;
  Begin
  If FForm <> Nil Then
    Begin
    FForm.Figure := Nil;
    FForm.Close;
    End;
  FDataList.Free;
  End;   {TomasCollectionFigure.Destroy}

Procedure TomasCollectionFigure.SetViewType(VType: CollectionViewType);
  Begin
  FView:=VType;
  End;   {TomasCollectionFigure.SetViewType}

Procedure TomasCollectionFigure.AddCollection(Collection: TomasCollectionData;
 Title: String);
  Var
   NewStatusBar: TCollectionStatusBar;
   NewPanel: TStatusPanel;
  Begin
  With Collection Do
    Begin
    FDataList.Add(Collection);
    FFigure:=Self;
    NewStatusBar:=TCollectionStatusBar.Create(FForm);
    Collection.FStatusBar:=NewStatusBar;
    NewStatusBar.Collection:=Collection;
    NewStatusBar.SimplePanel:=FALSE;
    NewStatusBar.SizeGrip:=FALSE;
    NewStatusBar.OnDrawPanel:=GoDrawPanel;
    NewPanel:=NewStatusBar.Panels.Add;
    NewPanel.Width:=30;
    NewPanel.Style:=psOwnerDraw;
    NewPanel:=NewStatusBar.Panels.Add;
    NewPanel.Width:=200;
    NewPanel.Text:=Title;
    NewPanel:=NewStatusBar.Panels.Add;
    NewPanel.Width:=80;
    NewPanel:=NewStatusBar.Panels.Add;
    NewPanel.Width:=80;
    NewPanel:=NewStatusBar.Panels.Add;
    NewPanel.Width:=80;
    NewPanel:=NewStatusBar.Panels.Add;
    NewPanel.Width:=80;
    NewPanel:=NewStatusBar.Panels.Add;
    NewPanel.Width:=80;
    NewStatusBar.Parent:=FForm;
    FForm.StatusBarList.Add(NewStatusBar);
    NewStatusBar.Show;
    NewStatusBar.BringToFront;
    If ViewType = Histogram Then
      Begin
      MakeHistograms;
      End;
    FForm.ReDrawFigure;
    RefreshData;
    Repaint;
    End;
  End;

Procedure TomasCollectionFigure.RemoveCollection(Collection: TomasCollectionData);
  Begin
  FDataList.Remove(Collection);
  FForm.StatusBarList.Remove(Collection.FStatusBar);
  Collection.FStatusBar.Free;
  Collection.FStatusBar:=Nil;
  Collection.FFigure:=Nil;
  If ViewType = Histogram Then
    begin
    MakeHistograms;
    End;
  End;

Function TomasCollectionFigure.NrOfCollections: Integer;
  Begin
  Result:=FDataList.Count;
  End;

Procedure TomasCollectionFigure.SetFValueAdded;
  Begin
  FValueAdded:=TRUE;
  End;

Function BarOrder(Item1,Item2: Pointer): Integer;
  Var
    Val1: PCollectionValue;
    Val2: PCollectionValue;
    R: Double;
  Begin
  With TomasCollectionData(Item1) Do
    Begin
    Val1:=FValues.Items[LastValueIndex];
    End;
  With TomasCollectionData(Item2) Do
    Begin
    Val2:=FValues.Items[LastValueIndex];
    End;
  R:=Val1^.XValue - Val2^.XValue;
  If R = 0 Then
    Begin
    R:=Val2^.Yvalue - Val1^.Yvalue;
    End;
  Result:=0;
  If R < 0 Then
    Begin
    Result:=-1;
    End;
  If R > 0 Then
    begin
    Result:=1;
    End;
  End;

Function HistDataSort(Item1,Item2: Pointer): Integer;
  Var
    R: Double;
  Begin
  R:=PHistDataLink(Item2)^.Percentage - PHistDataLink(Item1)^.Percentage;
  Result:=0;
  if R < 0 Then
    Begin
    Result:=-1;
    End;
  If R > 0 Then
    begin
    Result:=1;
    End;
  End;

Function TomasCollectionFigure.GetNextBarValue(Var Val: PCollectionValue): TomasCollectionData;
  Var
    Collection: TomasCollectionData;
    i: Integer;
    Ix: Integer;
    StartFound: Boolean;
  Begin
  If FFirstCall Then
    Begin
    FFirstCall:=FALSE;
    For i:=1 To FDataList.Count Do
      Begin
      Collection:=FDataList.Items[i-1];
      If Collection.FValues.Count > 0 Then
        Begin
        If FValueAdded Then
          begin
          Collection.LastValueIndex:=Collection.FValues.Count - 1;
          FValueAdded:=FALSE;
          FLastValues.Add(Collection);
          End
        Else
          Begin
          Ix:=0;
          StartFound:=FALSE;
          While (Ix<Collection.FValues.Count) And NOT(StartFound) Do
            Begin
            Val:=Collection.FValues.Items[Ix];
            If Val^.Xvalue < FMinMax.MinX Then
              Begin
              Inc(Ix);
              End
            Else
              Begin
              StartFound:=TRUE;
              End;
            End;
          If StartFound Then
            begin
            Collection.LastValueIndex:=Ix;
            FLastValues.Add(Collection);
            End;
          End;
        End;
      End;
    End;
  If FLastValues.Count = 0 Then
    Begin
    Result:=Nil;
    Val:=Nil;
    End
  Else
    Begin
    FLastValues.Sort(BarOrder);
    Collection:=FLastValues.First;
    Result:=Collection;
    With Collection Do
      Begin
      Val:=FValues.Items[LastValueIndex];
      Inc(LastValueIndex);
      If (LastValueIndex = FValues.Count) Or (Val^.XValue>Self.FMinMax.MaxX) Then
        Begin
        FLastValues.Remove(Collection);
        End;
      End;
    End;
  End;


Procedure TomasCollectionFigure.Repaint;
  Var
    Collection: TomasCollectionData;
    Value: PCollectionValue;
    i: Integer;
    j: integer;
    OrgColor: TColor;
    OrgBrush: TColor;
    XWidth: integer;
    YWidth: Integer;
    XRange: Double;
    YRange: Double;
    Y: Double;
    Total: Integer;
    BarWidth: Integer;
    Rect: TRect;
    HistData: PHistDataLink;
    HistList: TList;

  Begin
  {in case of bar paint collections in same points highest value first}
  {in case of graph there is no preferred order of drawing}
  OrgColor:=FForm.Canvas.Pen.Color;
  XWidth:=FForm.FRect.Right - FForm.FRect.Left;
  XRange:=FMinmax.MaxX - FMinMax.MinX;
  YWidth:=FForm.FRect.Bottom - FForm.FRect.Top;
  YRange:=FMinmax.MaxY - FMinMax.MinY;
  Case FView of
    Histogram:
      Begin
      HistList:=TList.Create;
      OrgBrush:=FForm.Canvas.Brush.Color;
      If FnrClasses <> 0 Then
        Begin
        BarWidth:=Trunc(XWidth/FNrClasses);
        End
      Else
        Begin
        BarWidth:=2;
        End;  
      If BarWidth <= 1 Then
        Begin
        BarWidth:=2;
        End;
      FForm.Canvas.Brush.Color:=OrgBrush;
      FForm.Canvas.Pen.Color:=OrgBrush;
      For i:=1 To FNrClasses Do   {clear histogram image}
        Begin
        Rect.Left:=1 + FForm.FRect.Left + (i-1) * BarWidth;
        Rect.Right := Rect.Left + BarWidth-1;
        Rect.Bottom:=FForm.FRect.Bottom;
        Rect.Top:=FForm.FRect.Top-1;
        FForm.Canvas.FillRect(Rect);
        End;
      If Ord(FHistogramView) mod 2 = 0 Then  {discrete included}
        Begin
        For i:=1 To FNrClasses Do
          Begin
          For j:=0 To FDataList.Count-1 Do
            Begin
            Collection:=FDataList.Items[j];
            If Collection.NrOfValues > 0 Then
              Begin
              HistData:=Collection.FClasses[i];
              HistData^.Percentage:=HistData^.NrObservations/Collection.NrOfValues;
              HistData^.Color:=Collection.CColor;
              HistList.Add(HistData);
              End;
            End;
          HistList.Sort(HistDataSort);
          FForm.Canvas.Pen.Color:=ClBlack;
          Rect.Left:=1 + FForm.FRect.Left + (i-1) * BarWidth;
          Rect.Right := Rect.Left + BarWidth-1;
          Rect.Bottom:=FForm.FRect.Bottom;
          While HistList.Count > 0 Do
            Begin
            HistData:=HistList.First;
            HistList.Remove(HistData);
            FForm.Canvas.Brush.Color:=HistData^.Color;
            Rect.Top:=FForm.FRect.Bottom - Round(YWidth * HistData^.Percentage);
            FForm.Canvas.Rectangle(Rect);
            End;
          End;
        End;
      If FHistogramView > HDiscrete Then
        Begin
        For i:=1 To FDataList.Count Do
          Begin
          Collection:=FDatalist.Items[i-1];
          If Collection.NrOfValues > 0 Then
            begin
            Total:=0;
            For j := 1 To Collection.FClasses.Count - 2 Do
              Begin
              HistData:=Collection.FClasses[j];
              Total:=Total + HistData^.NrObservations;
              Rect.Left:=1 + FForm.FRect.Left + (j-1) * BarWidth;
              Rect.Right := Rect.Left + BarWidth-1;
              Y:=Total / Collection.NrOfValues;
              Rect.Top:=FForm.FRect.Bottom - Round(YWidth * Y) - 1;
              Rect.Bottom:=Rect.Top + 2;
              FForm.Canvas.Pen.Color:=Collection.CColor;
              FForm.Canvas.Rectangle(Rect);
              End;
            End;
          End;
        End;
      FForm.Canvas.Brush.Color:=OrgBrush;
      HistList.Free;
      End;
    Bar:
      Begin
      OrgBrush:=FForm.Canvas.Brush.Color;
      If XRange <> 0 Then
        Begin
        BarWidth:=Trunc(XWidth/XRange);
        End
      Else
        Begin
        BarWidth:=2;
        ENd;
      If BarWidth <= 1 Then
        Begin
        BarWidth:=2;
        End;
      FFirstCall:=TRUE;
      Collection:=GetNextBarValue(Value);
      While Collection <> Nil Do {If Nil then Finished}
        Begin
        If XRange <> 0 Then
          Begin
          Rect.Left:=FForm.FRect.Left + Round(XWidth * ((Value^.XValue - FMinmax.MinX)/XRange))
           - BarWidth div 2;
          Rect.Right:=FForm.FRect.Left + Round(XWidth * ((Value^.XValue - FMinmax.MinX)/XRange))
           + BarWidth div 2;
          End
        Else
          Begin
          Rect.Left:=FForm.FRect.Left;
          Rect.Right:=FForm.FRect.Left + BarWidth;
          End;
        Rect.Bottom:=FForm.FRect.Bottom;
        If YRange <> 0 Then
          Begin
          Rect.Top:=FForm.FRect.Bottom - Round(YWidth * ((Value^.YValue - FMinMax.MinY)/YRange));
          End
        Else
          Begin
          Rect.Top:=FForm.FRect.Bottom;
          End;
        FForm.Canvas.Brush.Color:=Collection.CColor;
        FForm.Canvas.Pen.Color:=Collection.CColor;
        FForm.Canvas.Rectangle(Rect);
        Collection:=GetNextBarValue(Value); {sets XNow and LastX}
        End;
      FForm.Canvas.Brush.Color:=OrgBrush;
      End;
    Graph:
      Begin
      For i:=1 To FDataList.Count Do
        Begin
        Collection:=FDatalist.Items[i-1];
        FForm.Canvas.Pen.Color:=Collection.CColor;
        If FValueAdded  And (Collection.FValues.Count > 0) Then
          Begin
          J:=Collection.FValues.Count - 1;
          FValueAdded:=FALSE;
          End
        Else
          Begin
          Collection.XFrom:=-1;
          Collection.YFrom:=-1;
          J:=0;
          End;
        While J < Collection.FValues.Count Do
          Begin
          Value:=Collection.FValues.Items[j];
          If (Value^.XValue > FMinmax.MinX) And (Value^.XValue <= FMinmax.MaxX) Then
            Begin
            Collection.Xto:=FForm.FRect.Left + Round(XWidth * ((Value^.XValue - FMinmax.MinX)/XRange));
            Collection.Yto:=FForm.FRect.Bottom - Round(YWidth * ((Value^.YValue - FMinMax.MinY)/YRange));
            If Collection.XFrom <> -1 Then
              Begin
              FForm.Canvas.MoveTo(Collection.XFrom,Collection.YFrom);
              FForm.Canvas.LineTo(Collection.Xto,Collection.Yto);
              End;
            Collection.XFrom:=Collection.Xto;
            Collection.YFrom:=Collection.Yto;
            End;
          Inc(J);
          End;
        End;
      End;
    End;
  FForm.Canvas.Pen.Color:=OrgColor;
  End;

procedure TomasCollectionfigure.GoDrawPanel(StatusBar: TStatusBar;
 Panel: TStatusPanel; const Rect: TRect);
  Begin
  With StatusBar as TCollectionStatusBar Do
    Begin
    Canvas.Brush.Color := Collection.CColor;
    Canvas.FillRect(Rect);
    End;
  End;

Procedure TomasCollectionFigure.MakeHistograms;
  Var
    Collection: TomasCollectionData;
    i: Integer;
    j: Integer;
    HistData: PHistDataLink;
  Begin
  For i:=0 To FDataList.Count-1 Do
    Begin
    Collection:=FDataList.Items[i];
    While Collection.FClasses.Count > 0 Do
      Begin
      HistData:=Collection.FClasses.Items[0];
      Collection.FClasses.Remove(HistData);
      Dispose(HistData);
      End;
    New(HistData);
    HistData^.ClassLB:=-Never;
    HistData^.NrObservations:=0;
    Collection.FClasses.Add(HistData);
    For j:=1 To FNrClasses Do
      Begin
      New(HistData);
      HistData^.ClassLB:=FMinmax.MinY + (j-1) * FClassWidth;
      HistData^.NrObservations:=0;
      Collection.FClasses.Add(HistData);
      End;
    New(HistData);
    HistData^.ClassLB:=FMinmax.MinY + FNrClasses * FClassWidth;
    HistData^.NrObservations:=0;
    Collection.FClasses.Add(HistData);
    Collection.FillHistogram;
    End;
  End;

procedure TomasCollectionFigure.SaveToSettingsFile(Var F: TextFile);
  Begin
  WriteLn(F,'Title=' + Title);
  WriteLn(F,'Top=' + IntToStr(FForm.Top));
  WriteLn(F,'Left=' + IntToStr(FForm.Left));
  WriteLn(F,'Height=' + IntToStr(FForm.Height));
  WriteLn(F,'Width=' + IntToStr(FForm.Width));
  WriteLn(F,'Type=' + ViewTypeToString(ViewType));
  WriteLn(F,'MinX=' + FormatFloat('0.0',FMinMax.MinX));
  WriteLn(F,'MinY=' + FormatFloat('0.0',FMinMax.MinY));
  WriteLn(F,'MaxX=' + FormatFloat('0.0',FMinMax.MaxX));
  WriteLn(F,'MaxY=' + FormatFloat('0.0',FMinMax.MaxY));
  WriteLn(F,'Range=' + FormatFloat('0.0',FRange));
  WriteLn(F,'Classes=' + IntToStr(FNrClasses));
  WriteLn(F,'ClassWidth=' + FormatFloat('0.0',FClassWidth));
  WriteLn(F,'HistogramView=' + IntToStr(Ord(FHistogramView)));
  End;

Procedure TomasCollectionFigure.ReadFromSettingsFile(Var F: textFile);

  Function GetText: String;
    Var
     S: String;
    Begin
    ReadLn(F,S);
    Result:=Trim(Copy(S,Pos('=',S)+1,Length(S)-Pos('=',S)));
    End;

  Begin
  FForm.Top:=StrToInt(GetText);
  FForm.Left:=StrToInt(GetText);
  FForm.Height:=StrToInt(GetText);
  FForm.Width:=StrToInt(GetText);
  ViewType:=StringToViewType(GetText);
  FMinmax.MinX:=StrToFloat(GetText);
  FMinmax.MinY:=StrToFLoat(GetText);
  FMinmax.MaxX:=StrToFloat(GetText);
  FMinmax.MaxY:=StrToFloat(GetText);
  FRange:=StrToFloat(GetText);
  FNrClasses:=StrToInt(GetText);
  FClassWidth:=StrToFloat(GetText);
  FHistogramView:=HistogramViewType(StrToInt(GetText));
  End;



(***********************************************************************)
(*                     Methods of TCollectionForm                      *)
(***********************************************************************)



procedure TCollectionForm.GoCreateCollectionForm(Sender: TObject);
begin
Figure:=Nil;
StatusBarList:=TList.Create;
FRect.Top:=10;
FRect.Left:=50;
FRect.Bottom:=ClientHeight - 30;
FRect.Right:=ClientWidth - 10;
end;


procedure TCollectionForm.SetBarType(Sender: TObject);
begin
Figure.SetViewType(Bar);
RedrawFigure;
end;

procedure TCollectionForm.SetGraphType(Sender: TObject);
begin
Figure.SetViewType(Graph);
RedrawFigure;
end;

procedure TCollectionForm.SetHistogramType(Sender: TObject);
begin
Figure.SetViewType(Histogram);
Figure.MakeHistograms;
RedrawFigure;
end;



Procedure TCollectionForm.ReDrawFigure;
Var
  TheBar: TCollectionStatusBar;
  OldBrush: TColor;
  S: String;
  i: Integer;
  Step: Double;
  VStep: Double;
  Ypos: Integer;
  XPos: integer;
  Rect: TRect;
  nrSteps: Integer;
Begin
Rect.Top:=0;
Rect.Left:=0;
Rect.Bottom:=ClientHeight;
Rect.Right:=ClientWidth;
Canvas.FillRect(Rect);
If StatusBarList.Count > 0 Then
  Begin
  TheBar:=StatusBarList.Last;
  FRect.Bottom:=TheBar.Top - 30;
  FRect.Right:=ClientWidth - 10;
  OldBrush:=Canvas.Brush.Color;
  Canvas.Brush.Color:=clWhite;
  Canvas.FillRect(FRect);
  Canvas.Brush.Color:=OldBrush;
  If Figure.FView <> None Then
    Begin
    With Figure Do
      Begin
      Canvas.MoveTo(FRect.Left,FRect.Bottom);
      Canvas.LineTo(FRect.Right,FRect.Bottom);
      Canvas.MoveTo(FRect.Left,FRect.Top);
      Canvas.LineTo(FRect.Left,Frect.Bottom);
      Step:=(FRect.Bottom - Frect.Top + 1)/10.0;
      VStep:=0;
      If FView <> Histogram Then
        Begin
        VStep:=(FMinmax.MaxY - FMinmax.MinY)/10.0;
        End;
      For i := 0 To 10 Do
        Begin
        If FView = Histogram Then
          Begin
          S:=IntToStr(10 * i) + '%';
          End
        Else
          begin
          S:=FormatFloat('0.0',FMinmax.MinY + i * VStep);
          End;
        Ypos:=Trunc(FRect.Bottom - i * Step);
        Canvas.TextOut(FRect.Left - Canvas.TextWidth(S)-5,YPos-8,S);
        Canvas.MoveTo(FRect.Left - 4,YPos);
        Canvas.LineTo(FRect.Left,YPos);
        End;
      If FView<>Histogram Then
        Begin
        Step:=(FRect.Right - Frect.Left)/10.0;
        VStep:=(FMinmax.MaxX-FMinMax.MinX)/10.0;
        NrSteps:=10;
        End
      Else
        Begin
        If FNrClasses <> 0 Then
          begin
          Step:=(FRect.Right - Frect.Left)/FNrClasses;
          NrSteps:=FNrClasses;
          End
        Else
          Begin
          Step:=(FRect.Right - Frect.Left);
          NrSteps:=1;
          End;
        End;
      For i := 0 To NrSteps Do
        Begin
        If FView = Histogram Then
          Begin
          S:=FormatFloat('0.0',FMinmax.MinX + i * FClassWidth);
          End
        Else
          Begin
          S:=FormatFloat('0.0',FMinmax.MinX + i * VStep);
          End;
        XPos:=Trunc(FRect.Left + i * Step - (Canvas.TextWidth(S) div 2));
        Canvas.TextOut(Xpos,FRect.Bottom+3,S);
        XPos:=Trunc(Frect.Left + i * Step);
        Canvas.MoveTo(Xpos,FRect.Bottom);
        Canvas.LineTo(XPos,FRect.Bottom + 4);
        End;
      End;
    End;
  End;
If (Figure <> Nil) And (Figure.FView<> None) Then
  Begin
  Figure.Repaint;
  End;
End;

procedure TCollectionForm.AxisItemClick(Sender: TObject);
begin
If Figure.FView = Histogram Then
  Begin
  YMinEdit.Text:='0';
  YmaxEdit.Text:='100';
  YValuesGroup.Enabled:=FALSE;
  XRangeLabel.Caption:='#Classes';
  XMinEdit.Text:=FormatFloat('0.0',Figure.FMinMax.MinY);
  XRangeEdit.Text:=IntToStr(Figure.FNrClasses);
  ClassWidthEdit.Text:=FormatFloat('0.0',Figure.FClassWidth);
  HistogramTypeGroup.ItemIndex:=Ord(Figure.FHistogramView);
  HistogramTypeGroup.Show;
  ClassLabel.Show;
  ClassWidthEdit.Show;
  End
Else
  Begin
  YMinEdit.Text:=FormatFloat('0.0',Figure.FMinMax.MinY);
  YMaxEdit.Text:=FormatFloat('0.0',Figure.FMinMax.MaxY);
  YValuesGroup.Enabled:=TRUE;
  XRangeLabel.Caption:='Range';
  XMinEdit.Text:=FormatFloat('0.0',Figure.FMinMax.MinX);
  XRangeEdit.Text:=FormatFloat('0.0',Figure.FMinMax.MaxX - Figure.FMinMax.MinX);
  HistogramTypeGroup.Hide;
  ClassLabel.Hide;
  ClassWidthEdit.Hide;
  End;
XYPanel.Show;
XYPanel.BringToFront;
end;

procedure TCollectionForm.HideXYValues(Sender: TObject);
begin
XYPanel.Hide;
end;

procedure TCollectionForm.GoPaintCollectionForm(Sender: TObject);
begin
ReDrawFigure;
end;

procedure TCollectionForm.GoAcceptValues(Sender: TObject);
begin
With Figure Do
  Begin
  Try
    If FView = Histogram Then
      Begin
      FMinMax.MinY:=StrToFloat(XMinEdit.Text);
      FNrClasses:=StrToInt(XRangeEdit.Text);
      FClassWidth:=StrToFloat(ClassWidthEdit.Text);
      FHistogramView:=HistogramViewType(HistogramTypeGroup.ItemIndex);
      MakeHistograms;
      End
    Else
      Begin
      FMinMax.MinX:=StrToFloat(XMinEdit.Text);
      FMinmax.MaxX:=FMinMax.MinX + StrToFloat(XRangeEdit.Text);
      FMinmax.MinY:=StrToFloat(YMinEdit.Text);
      FMinMax.MaxY:=StrToFloat(YMaxEdit.Text);
      End;
    XYPanel.Hide;
    ReDrawFigure;
  Except
    On EConvertError Do
      MessageDlg('Invalid value',mtError,[mbOK],0);
    End;
  End;
end;

procedure TCollectionForm.GoResizeForm(Sender: TObject);
begin
RedrawFigure;
end;

procedure TCollectionForm.NoneItemClick(Sender: TObject);
begin
Figure.SetViewType(None);
RedrawFigure;
end;

procedure TCollectionForm.CheckCloseFigure(Sender: TObject;
  var Action: TCloseAction);
begin
If Figure = Nil Then
  Begin
  Action:=caFree;
  End
Else
  Begin
  Action:=caMinimize;
  End;
end;

procedure TCollectionForm.GoPrintFigure(Sender: TObject);
begin
If PrintDialog.Execute Then
  Begin
  Print;
  End;
end;

end.
